#include "scripts/spring.cpp"


void printkVector(kVec3&in inVec)
{
    Sys.Print("( " + inVec.x + ", " + inVec.y + ", " + inVec.z + " )");
}

float clamp(float x, float min, float max) //Built in function doesn't work?
{
    if (x > max) x = max;
    if (x < min) x = min;
    return x;
}

kVec3 getLookVector(float&in Pitch, float&in Yaw)
{
    float x,y,z;
    y = Math::Cos(Pitch) * Math::Cos(Yaw);
    x = Math::Cos(Pitch) * Math::Sin(Yaw);
    z = Math::Sin(-Pitch);
    return kVec3(x, y, z);
}

kQuat toQuaternion(float&in Yaw, float&in Pitch, float&in Roll)
{
    float cy = Math::Cos(Yaw * 0.5f);
    float sy = Math::Sin(Yaw * 0.5f);
    float cr = Math::Cos(Roll * 0.5f);
    float sr = Math::Sin(Roll * 0.5f);
    float cp = Math::Cos(Pitch * 0.5f);
    float sp = Math::Sin(Pitch * 0.5f);

    return kQuat(
            cy * sr * cp - sy * cr * sp,
            cy * cr * sp + sy * sr * cp, 
            sy * cr * cp - cy * sr * sp,
            cy * cr * cp + sy * sr * sp
        );
}

kQuat toQuaternion(kVec3&in eulerAngles)
{
    float Yaw = eulerAngles.x;
    float Pitch = eulerAngles.y;
    float Roll = eulerAngles.z;
    float cy = Math::Cos(Yaw * 0.5f);
    float sy = Math::Sin(Yaw * 0.5f);
    float cr = Math::Cos(Roll * 0.5f);
    float sr = Math::Sin(Roll * 0.5f);
    float cp = Math::Cos(Pitch * 0.5f);
    float sp = Math::Sin(Pitch * 0.5f);

    return kQuat(
            cy * sr * cp - sy * cr * sp,
            cy * cr * sp + sy * sr * cp, 
            sy * cr * cp - cy * sr * sp,
            cy * cr * cp + sy * sr * sp
        );
}

class TurokWeapon : ScriptObject
{
    kWeapon @self;
    
    TurokWeapon(kWeapon @actor)
    {
        @self = actor;
    }
    
    ~TurokWeapon() {}
    
    void OnTick(void) {}
    void OnSpawn(void) {}
    void OnBeginFire(void) {}
    void OnFire(void) {}
    void OnEndFire(void) {}
    void OnLower(void) {}
    void OnRaise(void) {}
    void OnHoldster(void) {}
    void OnFiredParticleFromAnim(kParticle@) {}
}

// weapon context
VecSpring rWeaponOffset;
VecSpring pWeaponOffset;
float m_fExtraSpinSpeed=0;
float m_fFireIntervals=0;
float m_fFireTime=0;
float m_fInitialYOffset=0;
float m_fInitialZOffset=0;
float m_fReturnTime=0;
float m_fScalar=1;
float m_fSpeed=0;
float m_fSpinAngle=0;
float m_fSpinRotVel=0; 
float m_fTimeBeforeAmmoConsumption=0;
float m_fYOffset=0;
float m_fZOffset=0;
float dx=0; float dy=0;
float lastx=0; float lasty=0;

int lastAnimID=-1;
bool m_bReleased=false;
bool m_bFiring=false;

class TurokSpringWeapon : TurokWeapon
{
    float weaponRefireTime = 0.9f;
    int32 actorId;
    bool regen = false;

    kActor@ wpnAtch;
    kVec3 pOffset;

    TurokSpringWeapon(kWeapon@ actor)
    {
        super(actor);
    }

    void regenerateActor()
    {
        @wpnAtch = ActorFactory.Spawn(actorId, kVec3(0,0,0),0,0,0);
        wpnAtch.RenderMeshComponent().Flags() |= RMCF_RENDER_AS_WEAPON;
        wpnAtch.Origin() = kVec3(0, 0, -1000);
        wpnAtch.Mark(false);
        regen = false;
    }

    void onSpringTick() 
    {
        if (regen)
            regenerateActor();

        if(@self.Owner().WeaponActor() == @self)
        {
            int currentAnimation;
            currentAnimation = self.AnimTrackComponent().PlayingID();
            if (currentAnimation != lastAnimID)
            {
                wpnAtch.AnimTrackComponent().Blend(currentAnimation, 0, 16, self.AnimTrackComponent().Flags());
                lastAnimID = currentAnimation;
            }

            if(self.AnimTrackComponent().Flags() != wpnAtch.AnimTrackComponent().Flags()) wpnAtch.AnimTrackComponent().Flags() = self.AnimTrackComponent().Flags();

            wpnAtch.Origin() = self.Origin();
            wpnAtch.Yaw() = self.Yaw();
            wpnAtch.Pitch() = self.Pitch();
            wpnAtch.Roll() = self.Roll();

            kVec3 originalEulerRotation = rWeaponOffset.getPosition(); originalEulerRotation.z=0;
            kVec3 rotation = rWeaponOffset.getPosition();
            kVec3 viewEuler = kVec3(self.Owner().Actor().Yaw(), self.Owner().Actor().Pitch(), self.Owner().Actor().Roll());
            float vVel = self.Owner().Actor().MovementComponent().Velocity().z;

            if (viewEuler.x >= 0 && lastx < 0)
            {
                dx = Math::Fabs(viewEuler.x - lastx); 
            }
            else if(viewEuler.x <= 0 && lastx > 0)
            {
                dx = -Math::Fabs(viewEuler.x - lastx);
            }
            else dx = viewEuler.x - lastx;
            dy = viewEuler.y - lasty;
            lastx = viewEuler.x;
            lasty = viewEuler.y;         

            rWeaponOffset.setTarget(kVec3(clamp(dx * 57.29578049f, -27, 27), (dy * 57.29578049f) + -vVel / 2, dy * 57.29578049f));
            pWeaponOffset.setTarget(kVec3(
                                rotation.x * 2,
                                rotation.y * 10,
                                (rotation.y * 10) + (-vVel)
                            ));
            pWeaponOffset.setVelocity(kVec3(
                                -rotation.x,
                                -rotation.y * 3,
                                rotation.y * 3
                            ));

            wpnAtch.Origin() = pOffset + pWeaponOffset.getPosition();
            wpnAtch.Yaw() = self.Yaw() + clamp(-rotation.x / 32, -45 * 0.0174532925f, 45 * 0.0174532925f);
            wpnAtch.Pitch() = -rotation.y / 32;
            wpnAtch.Roll() = clamp(rotation.z / 56, -12 * 0.0174532925f, 12 * 0.0174532925f);
        } else { wpnAtch.Origin() = kVec3(0, 0, -1000); }
    }

    void OnTick(void)
    { 
        onSpringTick();
	}

    void OnRaise(void)
    {
        m_fSpinRotVel = 0.0f;
        m_fSpeed = 4.0f;
        m_fFireTime = 0.0f;
        m_fReturnTime = 0.0f;
    }

    void OnLower(void)
    {
        m_fSpinRotVel = 0.0f;
        m_fSpeed = 4.0f;
        m_fFireTime = 0.0f;
        m_fReturnTime = 0.0f;
        wpnAtch.Origin() = kVec3(0, 0, -1000);
        self.StopLoopingSounds();
    }

    void OnFire(void)
    {
        if(self.AnimTrackComponent().TrackTime() >= weaponRefireTime)
        {
            self.AnimTrackComponent().Flags() |= ANF_CYCLECOMPLETED;
            wpnAtch.AnimTrackComponent().Blend(self.AnimTrackComponent().PlayingID(), 0, 16, self.AnimTrackComponent().Flags());
        }
    }

    void fireParticleFromAnimHitScan(kParticle& pBullet)
    {
        kVec3 vVel = pBullet.Velocity();
        kVec3 rOffset = rWeaponOffset.getPosition() * 0.0174532925f;
        kVec3 vDirection = getLookVector(self.Owner().Actor().Pitch() - rOffset.y, self.Owner().Actor().Yaw() - rOffset.x);
        float mag = vVel.Unit();

        pBullet.Direction() = vDirection;
        pBullet.Velocity() = vDirection * mag;    
    }

    void fireParticleFromAnim(kParticle& pBullet)
    {
        kVec3 vVel = pBullet.Velocity();
        kVec3 rOffset = rWeaponOffset.getPosition() * 0.0174532925f;
        kVec3 vDirection = getLookVector(self.Owner().Actor().Pitch() - rOffset.y, self.Owner().Actor().Yaw() - rOffset.x);
        float mag = vVel.Unit();
        
        pBullet.Direction() = vDirection;
        pBullet.Velocity() = vDirection * mag * m_fScalar;
    }

    void OnFiredParticleFromAnim(kParticle@ pBullet)
    {
        fireParticleFromAnimHitScan(pBullet);
    } 

    void OnSerialize(kDict& out dict)
    {
        //wpnAtch.Remove();
        //regen = true;
        
        SERIALIZE(m_fScalar);
        SERIALIZE(m_fSpinRotVel);
        SERIALIZE(m_fSpinAngle);
        SERIALIZE(m_bReleased);
        SERIALIZE(m_fZOffset);
        SERIALIZE(m_fInitialZOffset);
        SERIALIZE(m_fExtraSpinSpeed);
        SERIALIZE(m_fTimeBeforeAmmoConsumption);
        SERIALIZE(m_fFireIntervals);
        SERIALIZE(m_bFiring);
        SERIALIZE(m_fYOffset);
        SERIALIZE(m_fInitialYOffset);
        SERIALIZE(m_fReturnTime);
    }
    
    void OnDeserialize(kDict& in dict)
    {
        DESERIALIZE_FLOAT(m_fScalar);
        DESERIALIZE_FLOAT(m_fSpinRotVel);
        DESERIALIZE_FLOAT(m_fSpinAngle);
        DESERIALIZE_BOOL(m_bReleased);
        DESERIALIZE_FLOAT(m_fZOffset);
        DESERIALIZE_FLOAT(m_fInitialZOffset);
        DESERIALIZE_FLOAT(m_fExtraSpinSpeed);
        DESERIALIZE_FLOAT(m_fTimeBeforeAmmoConsumption);
        DESERIALIZE_FLOAT(m_fFireIntervals);
        DESERIALIZE_BOOL(m_bFiring);
        DESERIALIZE_FLOAT(m_fYOffset);
        DESERIALIZE_FLOAT(m_fInitialYOffset);
        DESERIALIZE_FLOAT(m_fReturnTime);
    }
}

final class TurokBow : TurokSpringWeapon
{
    TurokBow(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -45 , -100);
        actorId = kActor_Wpn_AttachBow;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
 
    void OnEndFire(void)
    {
        m_fScalar = Math::Max(0.5f, Math::Min(1.5f, self.AnimTrackComponent().PlayTime()));
    }   

    void OnFiredParticleFromAnim(kParticle@ pBullet) {
        fireParticleFromAnim(pBullet);
    }
}

final class TurokTekBow : TurokSpringWeapon
{   
    TurokTekBow(kWeapon @actor) 
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -45 , -100);
        actorId = kActor_Wpn_AttachTek;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
    
    void OnEndFire(void)
    {
        m_fScalar = Math::Max(0.5f, Math::Min(2.0f, self.AnimTrackComponent().PlayTime()));
    }

    void OnFiredParticleFromAnim(kParticle@ pBullet)
    {
        if(self.ScopeReady())
            return;

        fireParticleFromAnim(pBullet);
    }
}

final class TurokPistol : TurokSpringWeapon
{   
    TurokPistol(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -100);
        weaponRefireTime = 0.8f;
        actorId = kActor_Wpn_AttachPistol;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
    }
}

final class TurokChargeDart : TurokSpringWeapon
{   
    TurokChargeDart(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -100);
        actorId = kActor_Wpn_AttachCharge;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }

    void OnBeginFire(void)
    {
        self.PlaySound("sounds/shaders/Charge Dart PowerUp.ksnd");
    }
    
    void OnEndFire(void)
    {
        self.AnimTrackComponent().Flags() |= ANF_CYCLECOMPLETED;
        
        self.PlaySound("sounds/shaders/Charge Dart Fire.ksnd");
        self.FireProjectile("particles/Charge_Dart.particle", kVec3(0, 0, 0));
        
        float scalar = self.AnimTrackComponent().PlayTime() * 2.75f;
        if(scalar > 6.0f)
        {
            scalar = 6.0f;
        }
        
        switch(int(scalar))
        {
        case 0:
            self.FireProjectile("particles/Charge_Dart_1.particle", kVec3(0, -10, 0));
            self.Owner().ConsumeAmmo(1);
            break;
        case 1:
            self.FireProjectile("particles/Charge_Dart_2.particle", kVec3(0, -10, 0));
            self.Owner().ConsumeAmmo(1);
            break;
        case 2:
            self.FireProjectile("particles/Charge_Dart_3.particle", kVec3(0, -10, 0));
            self.Owner().ConsumeAmmo(1);
            break;
        case 3:
            self.FireProjectile("particles/Charge_Dart_4.particle", kVec3(0, -10, 0));
            self.Owner().ConsumeAmmo(1);
            break;
        case 4:
            self.FireProjectile("particles/Charge_Dart_5.particle", kVec3(0, -10, 0));
            self.Owner().ConsumeAmmo(1);
            break;
        case 5:
            self.FireProjectile("particles/Charge_Dart_6.particle", kVec3(0, -10, 0));
            self.Owner().ConsumeAmmo(2);
            break;
        case 6:
            self.FireProjectile("particles/Charge_Dart_7.particle", kVec3(0, -10, 0));
            self.Owner().ConsumeAmmo(3);
            break;
        }
    }
}

final class TurokShotgun : TurokSpringWeapon
{    
    TurokShotgun(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-50, -70 , -110);

        m_fInitialYOffset = self.Origin().y;
        m_fYOffset = 0.0f;
        m_fReturnTime = 0.0f;
        weaponRefireTime = 0.8f;
        actorId = kActor_Wpn_AttachShot;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
    
    void OnDeserialize(kDict& in dict)
    {
        DESERIALIZE_FLOAT(m_fYOffset);
        DESERIALIZE_FLOAT(m_fInitialYOffset);
        DESERIALIZE_FLOAT(m_fReturnTime);
        self.Origin().y = m_fYOffset;
    }

    void OnBeginFire(void)
    {
        m_fYOffset = 5*GAME_SCALE;
        m_fReturnTime = 0.0f;
    }
}

final class TurokPlasmaRifle : TurokSpringWeapon
{   
    TurokPlasmaRifle(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -110);
        m_fSpeed = 4.0f;
        m_fFireTime = 0.0f;
        weaponRefireTime = 0.8f;
        actorId = kActor_Wpn_AttachPlasma;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
    
    void OnBeginFire(void)
    {
        m_fSpeed = 4.0f;
        m_fFireTime = 0.0f;
    }
}

class TurokFireStormCannon : TurokSpringWeapon
{   
    Spring spinAngle;
    
    TurokFireStormCannon(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        spinAngle = Spring(14, 0.25f);
        pOffset = kVec3(-30, -100 , -30);
        m_fFireIntervals = 0.0f;
        m_fSpinRotVel = 0.0f;
        m_fSpinAngle = 0.0f;
        m_bFiring = false;
        actorId = kActor_Wpn_AttachFire;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
    
    void OnBeginFire(void)
    {
        m_fFireIntervals = 0.125f;
        self.PlayLoopingSound("sounds/shaders/Firestorm Cannon Spin.ksnd");
    }
    
    void OnFire(void)
    {
        m_fSpinRotVel = Math::IncMax(m_fSpinRotVel, 0.2f, 5);
        m_fSpinAngle += m_fSpinRotVel;
        
        m_bFiring = true;
        
        if(m_fSpinRotVel < 3.75f)
            return;
        
        m_fFireIntervals -= GAME_DELTA_TIME;
        if(m_fFireIntervals <= 0)
        {
            if(IsMultiplayer())
            {
                // Consume extra ammo in multiplayer
                self.Owner().ConsumeAmmo(1);
            }
            self.FireProjectile("particles/Firestorm_Cannon.particle", kVec3(0, 0, 6));
            self.PlayLoopingSound("sounds/shaders/Firestorm Cannon Fire.ksnd");
            m_fFireIntervals = 0.125f;
        }
    }
    
    void OnEndFire(void)
    {
        self.AnimTrackComponent().Flags() |= ANF_CYCLECOMPLETED;
        self.PlaySound("sounds/shaders/Firestorm Cannon Spin Stop.ksnd");
        self.StopLoopingSounds();
    }
    
    void OnTick(void)
    {
        spinAngle.setTarget(m_fSpinAngle);

        onSpringTick();

        wpnAtch.RenderMeshComponent().SetRotationOffset(1, Math::Deg2Rad(spinAngle.getPosition()), 0, 1, 0);
        wpnAtch.RenderMeshComponent().SetRotationOffset(7, Math::Deg2Rad(-spinAngle.getPosition()), 0, 1, 0);
        wpnAtch.RenderMeshComponent().SetRotationOffset(8, Math::Deg2Rad(spinAngle.getPosition()), 0, 1, 0);
        wpnAtch.RenderMeshComponent().SetRotationOffset(9, Math::Deg2Rad(spinAngle.getPosition()), 0, 1, 0);
        wpnAtch.RenderMeshComponent().SetRotationOffset(10, Math::Deg2Rad(spinAngle.getPosition()), 0, 1, 0);
        wpnAtch.RenderMeshComponent().SetRotationOffset(11, Math::Deg2Rad(spinAngle.getPosition()), 0, 1, 0);
        wpnAtch.RenderMeshComponent().SetRotationOffset(12, Math::Deg2Rad(spinAngle.getPosition()), 0, 1, 0);
        
        m_fSpinAngle += m_fSpinRotVel;
        
        if(m_bFiring)
        {
            m_bFiring = false;
            return;
        }
        
        m_fSpinRotVel = Math::IncMax(m_fSpinRotVel, Math::Deg2Rad(0.375f) * GAME_FRAME_UNIT, 0);
    }
}

final class TurokFireStormCannonMP : TurokFireStormCannon
{
    TurokFireStormCannonMP(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -110);

        actorId = kActor_Wpn_AttachFireMP;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
    }

    void OnFire(void)
    {
        m_fSpinRotVel = Math::IncMax(m_fSpinRotVel, 0.2f, 5);
        m_fSpinAngle += m_fSpinRotVel;
        
        m_bFiring = true;
        
        if(m_fSpinRotVel < 3.75f)
        {
            return;
        }
        
        m_fFireIntervals -= GAME_DELTA_TIME;
        if(m_fFireIntervals <= 0)
        {
            self.FireProjectile("particles/Multiplayer5.particle", kVec3(0, 0, 6));
            self.PlayLoopingSound("sounds/shaders/Firestorm Cannon Fire.ksnd");
            m_fFireIntervals = 0.125f;
        }
    }
    
    void OnTick(void)
    {
        onSpringTick();

        wpnAtch.RenderMeshComponent().SetRotationOffset(2, Math::Deg2Rad(m_fSpinAngle), 0, 1, 0);
        m_fSpinAngle += m_fSpinRotVel;
        
        if(m_bFiring)
        {
            m_bFiring = false;
            return;
        }
        
        m_fSpinRotVel = Math::IncMax(m_fSpinRotVel, Math::Deg2Rad(0.375f) * GAME_FRAME_UNIT, 0);
    }
}

final class TurokSunfirePod : TurokSpringWeapon
{   
    TurokSunfirePod(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -120);
        actorId = kActor_Wpn_AttachPod;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
    
    void OnEndFire(void)
    {
        m_fScalar = Math::Max(0.5f, Math::Min(1.5f, self.AnimTrackComponent().PlayTime()));
    }
    
    void OnFiredParticleFromAnim(kParticle@ pBullet)
    {
        fireParticleFromAnim(pBullet);
        self.Owner().ConsumeAmmo(1);
    }
}

final class TurokBore : TurokSpringWeapon
{   
    TurokBore(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-70, -70 , -100);
        actorId = kActor_Wpn_AttachBore;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
    
    void OnFiredParticleFromAnim(kParticle@ pBullet)
    {
        // target is set if this weapon is able to track things
        if(self.GetTarget() is null)
        {
            return;
        }

        kVec3 vVel = pBullet.Velocity();
        kVec3 rOffset = rWeaponOffset.getPosition() * 0.0174532925f;
        kVec3 vDirection = getLookVector(self.Owner().Actor().Pitch() - rOffset.y, self.Owner().Actor().Yaw() - rOffset.x);
        float mag = vVel.Unit();
        
        pBullet.Direction() = vDirection;
        pBullet.Velocity() = vDirection * mag;
        
        // have missile target the tracked object
        pBullet.SetTarget(self.GetTarget());
        
        // spawn a controller for manipulating the missile
        kActor@ pController =
        ActorFactory.Spawn(
            kActor_Misc_BoreController,
            pBullet.Origin(),
            0.0f,
            0.0f,
            0.0f);
            
        if(pController is null)
        {
            return;
        }
        
        // pController.PlayLoopingSound("sounds/shaders/Cerebral Bore Track Loop.ksnd");
        
        // make controller target the missile so it can control it
        pController.SetTarget(pBullet);
        
        // push the controller over the network for client replication
        if(Network.IsNetgame() && Network.IsServer() && !(pBullet.GetTarget() is null))
        {
            NetSend::Clear();
            
            // writes NPN_SCRIPT and then the string of the global function decl
            NetSend::WriteScriptHeader("void NetworkRecvSpawnBoreController(void)");
            NetSend::WriteInt16(pBullet.NetID());
            NetSend::WriteInt16(pBullet.GetTarget().NetID());
            
            // transmit pbuffer
            // when client recieves this, it parses the string which is the decl to the global function and then calls it
            Network.TransmitToAll();
        }
    }
}

final class TurokQuadMissileLauncher : TurokSpringWeapon
{   
    TurokQuadMissileLauncher(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -100);
        weaponRefireTime = 0.95f;
        actorId = kActor_Wpn_AttachMissile;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
    
    void OnBeginFire(void)
    {
        self.Owner().ConsumeAmmo(1);
    }
    
    void OnFiredParticleFromAnim(kParticle@ pBullet)
    {
        self.PlaySound("sounds/shaders/Scorpion Missile Launch.ksnd");
        
        // target is set if this weapon is able to track things
        if(self.GetTarget() is null)
        {
            return;
        }
        
        kVec3 vVel = pBullet.Velocity();
        kVec3 rOffset = rWeaponOffset.getPosition() * 0.0174532925f;
        kVec3 vDirection = getLookVector(self.Owner().Actor().Pitch() - rOffset.y, self.Owner().Actor().Yaw() - rOffset.x);
        float mag = vVel.Unit();

        
        pBullet.Direction() = vDirection;
        pBullet.Velocity() = vDirection * mag;

        // have missile target the tracked object
        pBullet.SetTarget(self.GetTarget());
        
        // spawn a controller for manipulating the missile
        kActor@ pController =
        ActorFactory.Spawn(
            kActor_Misc_MissileController,
            pBullet.Origin(),
            0.0f,
            0.0f,
            0.0f);
            
        if(pController is null)
        {
            return;
        }
        
        // make controller target the missile so it can control it
        pController.SetTarget(pBullet);
    }
}

final class TurokFlameThrower : TurokSpringWeapon
{
    TurokFlameThrower(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(60, 60 , -20);
        actorId = kActor_Wpn_AttachFlame;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
    
    void OnBeginFire(void)
    {
        self.StopLoopingSounds();
        self.PlayLoopingSound("sounds/shaders/Flame Thrower Loop.ksnd");
    }
    
    void OnEndFire(void)
    {
        self.AnimTrackComponent().Flags() |= ANF_CYCLECOMPLETED;
        
        self.StopLoopingSounds();
        self.PlayLoopingSound("sounds/shaders/Flame Thrower Gas Loop.ksnd");
    }
    
    void OnFire(void)
    {
        const float fAmount = IsMultiplayer() ? 12 : 6;
        
        m_fTimeBeforeAmmoConsumption += (GAME_DELTA_TIME*fAmount);
        
        if(m_fTimeBeforeAmmoConsumption >= 1.0f)
        {
            self.Owner().ConsumeAmmo(1);
            m_fTimeBeforeAmmoConsumption = 0;
        }
    }
    
    void OnRaise(void)
    {
        self.PlayLoopingSound("sounds/shaders/Flame Thrower Gas Loop.ksnd");
    }
}

final class TurokRazorWind : TurokSpringWeapon
{   
    TurokRazorWind(kWeapon @actor)
    {
        super(actor);
        m_fInitialZOffset = self.Origin().z;
        m_fZOffset = m_fInitialZOffset;
        m_fSpinAngle = 0.0;
        m_fExtraSpinSpeed = 0.0;
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -120);
        actorId = kActor_Wpn_AttachRazor;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }

    void OnTick(void)
    {
        onSpringTick();

        wpnAtch.RenderMeshComponent().SetRotationOffset(2, m_fSpinAngle, 0, 0, 1);
        m_fSpinAngle += Math::Deg2Rad(10.0f + (m_fExtraSpinSpeed * 20.0)) * 0.5f * GAME_FRAME_TIME;
        m_fExtraSpinSpeed = Math::Max(m_fExtraSpinSpeed - GAME_DELTA_TIME, 0.0f);
        
        self.Origin().y = m_fZOffset;
        
        if(m_fZOffset != m_fInitialZOffset)
        {
            m_fZOffset = (m_fInitialZOffset - m_fZOffset) * 0.1f + m_fZOffset;
        }
        
        if(self.GetTarget() is null)
        {
            return;
        }
        
        if(self.GetTarget().IsStale())
        {
            kActor@ pNull = null;
            
            self.SetTarget(pNull);
            wpnAtch.RenderMeshComponent().SwapOutMesh(0);
            self.LockControls(false);
            
            m_fZOffset = (5*GAME_SCALE);
            m_fExtraSpinSpeed = 1.0f;
        }
    }
    
    void OnFiredParticleFromAnim(kParticle@ pBullet)
    {
        wpnAtch.RenderMeshComponent().AltTexture() = -1;
        wpnAtch.RenderMeshComponent().SwapOutMesh(1);
        self.LockControls(true);
        
        // spawn a controller for manipulating the blade
        kActor@ pController =
        ActorFactory.Spawn(
            kActor_Misc_BladeController,
            pBullet.Origin(),
            0.0f,
            0.0f,
            0.0f);
            
        if(pController is null)
        {
            return;
        }
        
        kVec3 vVel = pBullet.Velocity();
        kVec3 rOffset = rWeaponOffset.getPosition() * 0.0174532925f;
        kVec3 vDirection = getLookVector(self.Owner().Actor().Pitch() - rOffset.y, self.Owner().Actor().Yaw() - rOffset.x);
        float mag = vVel.Unit();

        
        pBullet.Direction() = vDirection;
        pBullet.Velocity() = vDirection * mag;

        // target the controller so we can monitor it's lifetime
        self.SetTarget(pController);
        
        // make controller target the blade so it can control it
        pController.SetTarget(pBullet);
        pController.PlayLoopingSound("sounds/shaders/Razor Wind Fly Loop.ksnd");
    }
}

final class TurokTorpedoLauncher : TurokSpringWeapon
{   
    TurokTorpedoLauncher(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-80, -70 , -100);
        actorId = kActor_Wpn_AttachTorpedo;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
    
    void OnTick(void)
    {
        onSpringTick();

        float fAccel, fTarget;
        
        if((self.Owner().Buttons() & BC_FORWARD) != 0)
        {
            kVec3 vVel = self.Owner().Actor().MovementComponent().Velocity();
            float fSpeed = vVel.Unit();
            
            fTarget = Math::Deg2Rad(24);
            fAccel = fSpeed * 0.0525f;
            
            self.PlayLoopingSound("sounds/shaders/Torpedo Launcher Motor Loop.ksnd");
        }
        else
        {
            fAccel = Math::Deg2Rad(0.5f);
            fTarget = (m_fSpinRotVel > 0) ? Math::Deg2Rad(1) : Math::Deg2Rad(-1);
            self.StopLoopingSounds();
        }
        
        m_fSpinRotVel = Math::IncMax(m_fSpinRotVel, fAccel, fTarget);
        m_fSpinAngle += m_fSpinRotVel;
        
        wpnAtch.RenderMeshComponent().SetRotationOffset(2, m_fSpinAngle, 0, 1, 0);
        
        int animID = self.AnimTrackComponent().PlayingID();
        
        // firing animation pauses for a really long time and its annoying.
        // so force the animation to complete half-way through playing
        switch(animID)
        {
        case ANIM_WPN_ATK1:
        case ANIM_WPN_ATK2:
        case ANIM_WPN_ATK3:
        case ANIM_WPN_ATK4:
            if(self.AnimTrackComponent().TrackTime() >= 0.5f)
            {
                self.AnimTrackComponent().Flags() |= ANF_CYCLECOMPLETED;
                wpnAtch.AnimTrackComponent().Blend(self.AnimTrackComponent().PlayingID(), 0, 16, self.AnimTrackComponent().Flags());
            }
            break;
        default:
            break;
        };
    }
    
    void OnLower(void)
    {
        self.StopLoopingSounds();
        m_fSpinAngle = 0;
        wpnAtch.RenderMeshComponent().SetRotationOffset(2, Math::Deg2Rad(0), 0, 1, 0);
        wpnAtch.Origin() = kVec3(0, 0, -1000);
    }
    
    void OnFiredParticleFromAnim(kParticle@ pBullet)
    {
        // hack so that we consume ammo only once per fire animation
        if(self.AnimTrackComponent().TrackTime() < 0.3f)
        {
            self.Owner().ConsumeAmmo(1);
        }

        kVec3 vVel = pBullet.Velocity();
        kVec3 rOffset = rWeaponOffset.getPosition() * 0.0174532925f;
        kVec3 vDirection = getLookVector(self.Owner().Actor().Pitch() - rOffset.y, self.Owner().Actor().Yaw() - rOffset.x);
        float mag = vVel.Unit();
        
        pBullet.Direction() = vDirection;
        pBullet.Velocity() = vDirection * mag;
    }
}

final class TurokNuke : TurokSpringWeapon
{   
    TurokNuke(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -100);
        actorId = kActor_Wpn_AttachNuke;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
    
    void OnBeginFire(void)
    {
        self.PlayLoopingSound("sounds/shaders/Nuke Windup.ksnd");
    }

    void OnEndFire(void)
    {
        self.StopLoopingSounds();
        self.AnimTrackComponent().Flags() |= ANF_CYCLECOMPLETED;
        
        if(self.AnimTrackComponent().PlayTime() >= 2.0f)
        {
            m_fSpinRotVel = 2;
            wpnAtch.RenderMeshComponent().SetRotationOffset(1, Math::Deg2Rad(m_fSpinAngle), 0, 1, 0);
            wpnAtch.RenderMeshComponent().SetRotationOffset(12, Math::Deg2Rad(-m_fSpinAngle*2), 0, 1, 0);
            return;
        }
        
        m_bReleased = true;
        
        // note that this function is called before the engine sets the 'discharge'
        // animation on the weapon. flagging nointerrupt will prevent that from happening,
        // allowing us to set this animation
        self.AnimTrackComponent().Blend(ANIM_WPN_IDLE, 4, 16, ANF_NOINTERRUPT);
    }
    
    void OnTick(void)
    {
        onSpringTick();

        if(m_bReleased)
        {
            // clear the flag so we can immediately start uMath::Sing the weapon again
            self.AnimTrackComponent().Flags() |= ANF_CYCLECOMPLETED;
            self.AnimTrackComponent().Flags() &= ~ANF_NOINTERRUPT;
            m_bReleased = false;
        }
        
        int animID = self.AnimTrackComponent().PlayingID();
        
        if(animID == ANIM_WPN_CHARGE)
        {
            m_fSpinRotVel = Math::IncMax(m_fSpinRotVel, 0.025f, 8);
        }
        else if(animID == ANIM_WPN_IDLE || animID == ANIM_WPN_RUN)
        {
            m_fSpinRotVel = Math::IncMax(m_fSpinRotVel, Math::Deg2Rad(0.375f) * GAME_FRAME_UNIT, 2);
        }
        else
        {
            return;
        }
        
        wpnAtch.RenderMeshComponent().SetRotationOffset(1, Math::Deg2Rad(m_fSpinAngle), 0, 1, 0);
        wpnAtch.RenderMeshComponent().SetRotationOffset(12, Math::Deg2Rad(-m_fSpinAngle*2), 0, 1, 0);
        
        m_fSpinAngle += m_fSpinRotVel;
    }
    
    void OnLower(void)
    {
        m_fSpinAngle = 0;
        wpnAtch.RenderMeshComponent().SetRotationOffset(1, Math::Deg2Rad(0), 0, 1, 0);
        wpnAtch.RenderMeshComponent().SetRotationOffset(12, Math::Deg2Rad(0), 0, 1, 0);
        wpnAtch.Origin() = kVec3(0, 0, -1000);
    }
}

final class TurokBlade : TurokSpringWeapon
{
    TurokBlade(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -45 , -120);
        actorId = kActor_Wpn_AttachBlade;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokTalon : TurokSpringWeapon
{
    TurokTalon(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -55 , -110);
        actorId = kActor_Wpn_AttachTalon;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokMag : TurokSpringWeapon
{
    TurokMag(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -105);
        actorId = kActor_Wpn_AttachMag;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokShred : TurokSpringWeapon
{
    TurokShred(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-50, -70 , -115);
        actorId = kActor_Wpn_AttachShred;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokTranq : TurokSpringWeapon
{
    TurokTranq(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -115);
        actorId = kActor_Wpn_AttachTranq;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokPFM : TurokSpringWeapon
{
    TurokPFM(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(30, -80 , -30);
        actorId = kActor_Wpn_AttachPFM;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokGrenade : TurokSpringWeapon
{
    TurokGrenade(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -105);
        actorId = kActor_Wpn_AttachGrenade;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokHarpoon : TurokSpringWeapon
{
    TurokHarpoon(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -105);
        actorId = kActor_Wpn_AttachHarpoon;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokFlare : TurokSpringWeapon
{
    TurokFlare(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -105);
        actorId = kActor_Wpn_AttachFlare;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokCrossbow : TurokSpringWeapon
{
    TurokCrossbow(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -105);
        actorId = kActor_Wpn_AttachCrossbow;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokPlasmaMP : TurokSpringWeapon
{
    TurokPlasmaMP(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -105);
        actorId = kActor_Wpn_AttachPlasmaMP;
        
        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokRocketMP : TurokSpringWeapon
{
    TurokRocketMP(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -105);
        actorId = kActor_Wpn_AttachRocketMP;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokChargeMP : TurokSpringWeapon
{
    TurokChargeMP(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -105);
        actorId = kActor_Wpn_AttachChargeMP;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokAssaultMP : TurokSpringWeapon
{
    TurokAssaultMP(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -105);
        actorId = kActor_Wpn_AttachAssaultMP;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokNadeMP : TurokSpringWeapon
{
    TurokNadeMP(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -105);
        actorId = kActor_Wpn_AttachNadeMP;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokFirestormCannonMP : TurokSpringWeapon
{
    TurokFirestormCannonMP(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -105);
        actorId = kActor_Wpn_AttachFireStormMP;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}

final class TurokClaws : TurokSpringWeapon
{
    TurokClaws(kWeapon @actor)
    {
        super(actor);
        rWeaponOffset = VecSpring(3, 0.25f);
        pWeaponOffset = VecSpring(6, 0.25f);
        pOffset = kVec3(-60, -70 , -105);
        actorId = kActor_Wpn_AttachClaws;

        self.RenderMeshComponent().SetScale(0,kVec3(0,0,0));
        regenerateActor();
    }
}
